@@ -4,7 +4,7 @@
 # copyright (c) 2000, 2001, 2002 David Schweikert <dws@ee.ethz.ch>
 # released under the GNU General Public License
 
-######## Parse::Syslog 0.06 (automatically embedded) ########
+######## Parse::Syslog 1.02 (automatically embedded) ########
 package Parse::Syslog;
 use Carp;
 use Symbol;
@@ -21,23 +21,24 @@ my %months_map = (
     'jul' => 6, 'aug' => 7, 'sep' => 8,
     'oct' => 9, 'nov' =>10, 'dec' =>11,
 );
-# fast timelocal
-my $str2time_last_time;
-my $str2time_last_day;
-my $str2time_last_month;
-my $enable_year_decrement = 1; # year-increment algorithm: if in january, if december is seen, decrement
-                               # year
+# year-increment algorithm: if in january, if december is seen, decrement year
+my $enable_year_decrement = 1;
+# fast timelocal, cache minute's timestamp
+# don't cache more than minute because of daylight saving time switch
+my @str2time_last_minute;
+my $str2time_last_minute_timestamp;
 # 0: sec, 1: min, 2: h, 3: day, 4: month, 5: year
 sub str2time($$$$$$$)
 {
     my $GMT = pop @_;
-    my $day_secs = $_[2]*3600+$_[1]*60+$_[0];
-    if(defined $str2time_last_time) {
-        if( $_[3] == $str2time_last_day and
-            $_[4] == $str2time_last_month )
-        {
-            return $str2time_last_time + $day_secs;
-        }
+    if(defined $str2time_last_minute[4] and
+        $str2time_last_minute[0] == $_[1] and
+        $str2time_last_minute[1] == $_[2] and
+        $str2time_last_minute[2] == $_[3] and
+        $str2time_last_minute[3] == $_[4] and
+        $str2time_last_minute[4] == $_[5])
+    {
+        return $str2time_last_minute_timestamp + $_[0];
     }
     my $time;
     if($GMT) {
@@ -46,11 +47,22 @@ sub str2time($$$$$$$)
     else {
         $time = timelocal(@_);
     }
-    $str2time_last_time = $time - $day_secs;
-    $str2time_last_day = $_[3];
-    $str2time_last_month = $_[4];
+    @str2time_last_minute = @_[1..5];
+    $str2time_last_minute_timestamp = $time-$_[0];
     return $time;
 }
+sub _use_locale($)
+{
+    use POSIX qw(locale_h strftime);
+    my $old_locale = setlocale(LC_TIME);
+    for my $locale (@_) {
+        croak "new(): wrong 'locale' value: '$locale'" unless setlocale(LC_TIME, $locale);
+        for my $month (0..11) {
+            $months_map{strftime("%b", 0, 0, 0, 1, $month, 96)} = $month;
+        }
+    }
+    setlocale(LC_TIME, $old_locale);
+}
 sub new($$;%)
 {
     my ($class, $file, %data) = @_;
@@ -68,6 +80,17 @@ sub new($$;%)
         $data{file}=gensym;
         open($data{file}, "<$file") or croak "can't open $file: $!";
     }
+    if(defined $data{locale}) {
+        if(ref $data{locale} eq 'ARRAY') {
+            _use_locale @{$data{locale}};
+        }
+        elsif(ref $data{locale} eq '') {
+            _use_locale $data{locale};
+        }
+        else {
+            croak "'locale' parameter must be scalar or array of scalars";
+        }
+    }
     return bless \%data, $class;
 }
 sub _next_line($)
@@ -91,7 +114,7 @@ sub next($)
     line: while(my $str = $self->_next_line()) {
         # date, time and host 
         $str =~ /^
-            (\w{3})\s+(\d+)   # date  -- 1, 2
+            (\S{3})\s+(\d+)   # date  -- 1, 2
             \s
             (\d+):(\d+):(\d+) # time  -- 3, 4, 5
             \s
@@ -139,6 +162,9 @@ sub next($)
         # some systems send over the network their
         # hostname prefixed to the text. strip that.
         $text =~ s/^$host\s+//;
+        # discard ':' in HP-UX 'su' entries like this:
+        # Apr 24 19:09:40 remedy : su : + tty?? root-oracle
+        $text =~ s/^:\s+//;
         $text =~ /^
             ([^:]+?)        # program   -- 1
             (?:\[(\d+)\])?  # PID       -- 2
@@ -185,14 +211,19 @@ use RRDs;
 use strict;
 use File::Tail;
 use Getopt::Long;
+use POSIX 'setsid';
 
-my $VERSION = 0.20;
+my $VERSION = "1.2";
 
 # config
 my $rrdstep = 60;
 my $xpoints = 540;
 my $points_per_sample = 3;
 
+my $daemon_logfile = '/var/log/mailgraph.log';
+my $daemon_pidfile = '/var/run/mailgraph.pid';
+my $daemon_rrd_dir = '/var/log';
+
 # global variables
 my $logfile;
 my $rrd = "mailgraph.rrd";
@@ -201,8 +232,10 @@ my $year;
 my $this_minute;
 my %sum = ( sent => 0, received => 0, bounced => 0, rejected => 0, virus => 0, spam => 0 );
 my $rrd_inited=0;
+my $verbose=0;
 
 # prototypes
+sub daemonize();
 sub process_line($);
 sub event_sent($);
 sub event_received($);
@@ -216,11 +249,17 @@ sub update($);
 sub usage
 {
 	print "usage: mailgraph [*options*]\n\n";
+	print "  -h, --help         display this help and exit\n";
+	print "  -v, --verbose      be verbose about what you do\n";
+	print "  -V, --version      output version information and exit\n";
 	print "  -c, --cat          causes the logfile to be only read and not monitored\n";
 	print "  -l, --logfile f    monitor logfile f instead of /var/log/syslog\n";
-	print "  -h, --help         display this help and exit\n";
-	print "  -v, --version      output version information and exit\n";
 	print "  -y, --year         starting year of the log file (default: current year)\n";
+	print "      --host=HOST    use only entries for HOST (regexp) in syslog\n";
+	print "  -d, --daemon       start in the background\n";
+	print "  --daemon_pid=FILE  write PID to FILE instead of /var/run/mailgraph.pid\n";
+	print "  --daemon_rrd=DIR   write RRDs to DIR instead of /var/log\n";
+	print "  --daemon_log=FILE  write verbose-log to FILE instead of /var/log/mailgraph.log\n";
 
 	exit;
 }
@@ -228,8 +267,11 @@ sub usage
 sub main
 {
 	my %opt = ();
-	GetOptions(\%opt, 'help|h', 'cat|c', 'logfile|l=s', 'version|v', 'year|y=i')
-		or exit(1);
+	Getopt::Long::Configure('no_ignore_case');
+	GetOptions(\%opt, 'help|h', 'cat|c', 'logfile|l=s', 'version|V',
+		'year|y=i', 'host=s', 'verbose|v+', 'daemon|d!',
+		'daemon_pid=s', 'daemon_rrd=s', 'daemon_log=s'
+		) or exit(1);
 	usage if $opt{help};
 
 	if($opt{version}) {
@@ -237,6 +279,13 @@ sub main
 		exit;
 	}
 
+	$verbose = $opt{verbose} if $opt{verbose};
+
+	$daemon_pidfile = $opt{daemon_pid} if defined $opt{daemon_pid};
+	$daemon_logfile = $opt{daemon_log} if defined $opt{daemon_log};
+	$daemon_rrd_dir = $opt{daemon_rrd} if defined $opt{daemon_rrd};
+	daemonize if $opt{daemon};
+
 	my $logfile = defined $opt{logfile} ? $opt{logfile} : '/var/log/syslog';
 	my $file;
 	if($opt{cat}) {
@@ -245,10 +294,46 @@ sub main
 	else {
 		$file = File::Tail->new(name=>$logfile, tail=>-1);
 	}
-	my $parser = new Parse::Syslog($file, year => $opt{year});
-	while(my $sl = $parser->next) {
-		process_line($sl);
+	my $parser = new Parse::Syslog($file, year => $opt{year}, arrayref => 1);
+
+	if(not defined $opt{host}) {
+		while(my $sl = $parser->next) {
+			process_line($sl);
+		}
 	}
+	else {
+		my $host = qr/^$opt{host}$/i;
+		while(my $sl = $parser->next) {
+			process_line($sl) if $sl->[1] =~ $host;
+		}
+	}
+}
+
+sub daemonize()
+{
+	chdir $daemon_rrd_dir or die "mailgraph: can't chdir to $daemon_rrd_dir: $!";
+	-w $daemon_rrd_dir or die "mailgraph: can't write to $daemon_rrd_dir\n";
+	open STDIN, '/dev/null' or die "mailgraph: can't read /dev/null: $!";
+	if($verbose>0) {
+		open STDOUT, ">>$daemon_logfile"
+			or die "mailgraph: can't write to $daemon_logfile: $!";
+	}
+	else {
+		open STDOUT, '>/dev/null'
+			or die "mailgraph: can't write to /dev/null: $!";
+	}
+	defined(my $pid = fork) or die "mailgraph: can't fork: $!";
+	if($pid) {
+		# parent
+		open PIDFILE, ">$daemon_pidfile"
+			or die "mailgraph: can't write to $daemon_pidfile: $!\n";
+		print PIDFILE "$pid\n";
+		close(PIDFILE);
+		exit;
+	}
+	# child
+	setsid                  or die "mailgraph: can't start a new session: $!";
+	open STDERR, '>&STDOUT' or die "mailgraph: can't dup stdout: $!";
 }
 
 sub init_rrd($)
@@ -303,16 +388,20 @@ sub init_rrd($)
 sub process_line($)
 {
 	my $sl = shift;
-	my $time = $sl->{timestamp};
-	my $text = $sl->{text};
+	my $time = $sl->[0];
+	my $prog = $sl->[2];
+	my $text = $sl->[4];
 
-	if($sl->{program} =~ /^postfix\/(.*)/) {
+	if($prog =~ /^postfix\/(.*)/) {
 		my $prog = $1;
 		if($prog eq 'smtp') {
 			if($text =~ /\bstatus=sent\b/) {
-				event_sent($time);
+				# Do not count sending to Amavis
+				if($text !~ /\brelay=[^\s\[]*\[127\.0\.0\.1\]/) {
+					event_sent($time);
+				}
 			}
-			if($text =~ /\bstatus=bounced\b/) {
+			elsif($text =~ /\bstatus=bounced\b/) {
 				event_bounced($time);
 			}
 		}
@@ -322,29 +411,93 @@ sub process_line($)
 			}
 		}
 		elsif($prog eq 'smtpd') {
-			if($text =~ /^[0-9A-F]+: client=/) {
-				event_received($time);
+			if($text =~ /^[0-9A-F]+: client=(\S+)/) {
+				my $client = $1;
+				# do not count receiving from Amavis
+				if($client !~ /\[127\.0\.0\.1\]$/) {
+					event_received($time);
+				}
 			}
-			if($text =~ /^reject: /) {
+			elsif($text =~ /^([0-9A-F]+: )?reject: /) {
 				event_rejected($time);
 			}
 		}
 		elsif($prog eq 'cleanup') {
-			if($text =~ /^[0-9A-F]+: reject: /) {
+			if($text =~ /^[0-9A-F]+: (?:reject|discard): /) {
 				event_rejected($time);
 			}
 		}
+		elsif($prog eq 'pipe') {
+			if($text =~ /\bstatus=sent\b/) {
+				event_sent($time);
+			}
+		}
 	}
-	elsif($sl->{program} eq 'amavis') {
+	elsif($prog eq 'amavis') {
+		if($text =~ /^infected/) {
+			event_virus($time);
+		}
+		elsif($text =~ /^\(\d+-\d+\) infected/i) {
+			# amavisd-new 20021126
+			event_virus($time);
+		}
+		elsif($text =~ /^Virus found/) {
+			# AMaViS 0.3.12pre8
+			event_virus($time);
+		}
+		elsif($text =~ /^spam_scan: Yes/) {
+			event_spam($time);
+		}
+		elsif($text =~ /^\(\d+-\d+\) spam/i) {
+			# old amavisd: spam, new: SPAM
+			event_spam($time);
+		}
+	}
+	elsif($prog eq 'vagatefwd') {
+		# Vexira antivirus
+		if($text =~ /^VIRUS/) {
+			event_virus($time);
+		}
+	}
+	elsif($prog eq 'avgatefwd') {
+		# AntiVir MailGate
+		if($text =~ /^Alert!/) {
+			event_virus($time);
+		}
+		elsif($text =~ /blocked\.$/) {
+			event_virus($time);
+		}
+	}
+	elsif($prog eq 'avcheck') {
+		# avcheck
 		if($text =~ /^infected/) {
 			event_virus($time);
 		}
 	}
-	elsif($sl->{program} eq 'spamd') {
+	elsif($prog eq 'spamd') {
 		if($text =~ /^identified spam/) {
 			event_spam($time);
 		}
 	}
+	elsif($prog eq 'spamproxyd') {
+		if($text =~ /^\s*SPAM/ or $text =~ /^identified spam/) {
+			event_spam($time);
+		}
+	}
+	elsif($prog eq 'drweb-postfix') {
+		# DrWeb
+		if($text =~ /infected$/) {
+			event_virus($time);
+		}
+	}
+	elsif($prog eq 'BlackHole') {
+		if($text =~ /Virus/) {
+			event_virus($time);
+		}
+		if($text =~ /(?:RBL|Razor|Spam)/) {
+			event_spam($time);
+		}
+	}
 }
 
 sub event_sent($)
@@ -392,13 +545,12 @@ sub update($)
 	return 1 if $m == $this_minute;
 	return 0 if $m < $this_minute;
 
-	#print "$m\n";
-	print "update $this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}:$sum{virus}:$sum{spam}\n";
+	print "update $this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}:$sum{virus}:$sum{spam}\n" if $verbose > 0;
 	RRDs::update $rrd, "$this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}";
 	RRDs::update $rrd_virus, "$this_minute:$sum{virus}:$sum{spam}";
 	if($m > $this_minute+$rrdstep) {
 		for(my $sm=$this_minute+$rrdstep;$sm<$m;$sm+=$rrdstep) {
-			print "update $sm:0:0:0:0:0:0 (SKIP)\n";
+			print "update $sm:0:0:0:0:0:0 (SKIP)\n" if $verbose > 0;
 			RRDs::update $rrd, "$sm:0:0:0:0";
 			RRDs::update $rrd_virus, "$sm:0:0";
 		}
